// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

/******************************************************************

DO NOT MODIFY. AUTOGENERATED FILE.
This file is generated using the logic from <root>/src/scripts/genEventPipe.py

******************************************************************/

#include <eventpipe/ep-rt-config.h>
#ifdef ENABLE_PERFTRACING
#include <eventpipe/ep.h>
#include <eventpipe/ep-event.h>
#include "clreventpipewriteevents.h"

#ifndef W
#define W(str) L##str
#endif
#ifndef SL
#define SL(str) str
#endif
#ifndef ERROR_SUCCESS
#define ERROR_SUCCESS 0L
#endif
#ifndef ERROR_WRITE_FAULT
#define ERROR_WRITE_FAULT 29L
#endif

/*
 * Forward declares of functions.
 */

bool
resize_buffer (
    uint8_t **buffer,
    size_t *size,
    size_t current_size,
    size_t new_size,
    bool *fixed_buffer);

bool
write_buffer (
    const uint8_t *value,
    size_t value_size,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer);

bool
write_buffer_string_utf8_t (
    const ep_char8_t *value,
    size_t value_len,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer);

bool
write_buffer_string_utf8_to_utf16_t (
    const ep_char8_t *value,
    size_t value_len,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer);

static
inline
bool
write_buffer_guid_t (
    const uint8_t *value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    return write_buffer (value, EP_GUID_SIZE, buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_uint8_t (
    uint8_t value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    return write_buffer ((const uint8_t *)&value, sizeof (uint8_t), buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_uint16_t (
    uint16_t value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    value = ep_rt_val_uint16_t (value);
    return write_buffer ((const uint8_t *)&value, sizeof (uint16_t), buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_uint32_t (
    uint32_t value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    value = ep_rt_val_uint32_t (value);
    return write_buffer ((const uint8_t *)&value, sizeof (uint32_t), buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_int32_t (
    int32_t value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    value = ep_rt_val_int32_t (value);
    return write_buffer ((const uint8_t *)&value, sizeof (int32_t), buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_uint64_t (
    uint64_t value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    value = ep_rt_val_uint64_t (value);
    return write_buffer ((const uint8_t *)&value, sizeof (uint64_t), buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_int64_t (
    int64_t value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    value = ep_rt_val_int64_t (value);
    return write_buffer ((const uint8_t *)&value, sizeof (int64_t), buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_double_t (
    double value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
#if BIGENDIAN
    uint64_t value_as_uint64_t;
    memcpy (&value_as_uint64_t, &value, sizeof (uint64_t));
    value_as_uint64_t = ep_rt_val_uint64_t (value_as_uint64_t);
    memcpy (&value, &value_as_uint64_t, sizeof (uint64_t));
#endif
    return write_buffer ((const uint8_t *)&value, sizeof (double), buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_bool_t (
    bool value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    return write_buffer_int32_t (value, buffer, offset, size, fixed_buffer);
}

static
inline
bool
write_buffer_uintptr_t (
    uintptr_t value,
    uint8_t **buffer,
    size_t *offset,
    size_t *size,
    bool *fixed_buffer)
{
    value = ep_rt_val_uintptr_t (value);
    return write_buffer ((const uint8_t *)&value, sizeof (uintptr_t), buffer, offset, size, fixed_buffer);
}

static
inline
EventPipeEvent *
provider_add_event (
    EventPipeProvider *provider,
    uint32_t event_id,
    uint64_t keywords,
    uint32_t event_version,
    EventPipeEventLevel level,
    bool need_stack)
{
    return ep_provider_add_event (provider, event_id, keywords, event_version, level, need_stack, NULL, 0);
}

static
inline
EventPipeProvider *
create_provider (
    const wchar_t *provider_name,
    EventPipeCallback callback_func)
{
    ep_char8_t *provider_name_utf8 = NULL;

#if WCHAR_MAX == 0xFFFF
    provider_name_utf8 = g_utf16_to_utf8 ((const gunichar2 *)provider_name, -1, NULL, NULL, NULL);
#else
    provider_name_utf8 = g_ucs4_to_utf8 ((const gunichar *)provider_name, -1, NULL, NULL, NULL);
#endif

    ep_return_null_if_nok (provider_name_utf8 != NULL);

    EventPipeProvider *provider = ep_create_provider (provider_name_utf8, callback_func, NULL);

    g_free (provider_name_utf8);
    return provider;
}

const wchar_t* DotNETRuntimeName = W("Microsoft-Windows-DotNETRuntime");
EventPipeProvider *EventPipeProviderDotNETRuntime = NULL;
EventPipeEvent *EventPipeEventGCStart_V2 = NULL;
EventPipeEvent *EventPipeEventGCEnd_V1 = NULL;
EventPipeEvent *EventPipeEventBulkType = NULL;
EventPipeEvent *EventPipeEventGCBulkRootEdge = NULL;
EventPipeEvent *EventPipeEventGCBulkRootConditionalWeakTableElementEdge = NULL;
EventPipeEvent *EventPipeEventGCBulkNode = NULL;
EventPipeEvent *EventPipeEventGCBulkEdge = NULL;
EventPipeEvent *EventPipeEventGCBulkRootStaticVar = NULL;
EventPipeEvent *EventPipeEventThreadPoolWorkerThreadStart = NULL;
EventPipeEvent *EventPipeEventThreadPoolWorkerThreadStop = NULL;
EventPipeEvent *EventPipeEventThreadPoolWorkerThreadAdjustmentSample = NULL;
EventPipeEvent *EventPipeEventThreadPoolWorkerThreadAdjustmentAdjustment = NULL;
EventPipeEvent *EventPipeEventThreadPoolWorkerThreadAdjustmentStats = NULL;
EventPipeEvent *EventPipeEventThreadPoolWorkerThreadWait = NULL;
EventPipeEvent *EventPipeEventThreadPoolMinMaxThreads = NULL;
EventPipeEvent *EventPipeEventThreadPoolWorkingThreadCount = NULL;
EventPipeEvent *EventPipeEventThreadPoolIOEnqueue = NULL;
EventPipeEvent *EventPipeEventThreadPoolIODequeue = NULL;
EventPipeEvent *EventPipeEventThreadPoolIOPack = NULL;
EventPipeEvent *EventPipeEventMethodDetails = NULL;
EventPipeEvent *EventPipeEventTypeLoadStart = NULL;
EventPipeEvent *EventPipeEventTypeLoadStop = NULL;
EventPipeEvent *EventPipeEventExceptionThrown_V1 = NULL;
EventPipeEvent *EventPipeEventExceptionCatchStart = NULL;
EventPipeEvent *EventPipeEventExceptionCatchStop = NULL;
EventPipeEvent *EventPipeEventExceptionFinallyStart = NULL;
EventPipeEvent *EventPipeEventExceptionFinallyStop = NULL;
EventPipeEvent *EventPipeEventExceptionFilterStart = NULL;
EventPipeEvent *EventPipeEventExceptionFilterStop = NULL;
EventPipeEvent *EventPipeEventExceptionThrownStop = NULL;
EventPipeEvent *EventPipeEventContentionStart_V1 = NULL;
EventPipeEvent *EventPipeEventContentionStart_V2 = NULL;
EventPipeEvent *EventPipeEventContentionStop = NULL;
EventPipeEvent *EventPipeEventContentionStop_V1 = NULL;
EventPipeEvent *EventPipeEventContentionLockCreated = NULL;
EventPipeEvent *EventPipeEventThreadCreated = NULL;
EventPipeEvent *EventPipeEventThreadTerminated = NULL;
EventPipeEvent *EventPipeEventMethodLoad_V1 = NULL;
EventPipeEvent *EventPipeEventMethodLoadVerbose_V1 = NULL;
EventPipeEvent *EventPipeEventMethodJittingStarted_V1 = NULL;
EventPipeEvent *EventPipeEventMethodJitMemoryAllocatedForCode = NULL;
EventPipeEvent *EventPipeEventMethodILToNativeMap = NULL;
EventPipeEvent *EventPipeEventDomainModuleLoad_V1 = NULL;
EventPipeEvent *EventPipeEventModuleLoad_V2 = NULL;
EventPipeEvent *EventPipeEventModuleUnload_V2 = NULL;
EventPipeEvent *EventPipeEventAssemblyLoad_V1 = NULL;
EventPipeEvent *EventPipeEventAssemblyUnload_V1 = NULL;

bool EventPipeEventEnabledGCStart_V2(void)
{
    return ep_event_is_enabled(EventPipeEventGCStart_V2);
}

uint32_t EventPipeWriteEventGCStart_V2(
    const uint32_t Count,
    const uint32_t Depth,
    const uint32_t Reason,
    const uint32_t Type,
    const uint16_t ClrInstanceID,
    const uint64_t ClientSequenceNumber,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledGCStart_V2())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(Count, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Depth, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Reason, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Type, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ClientSequenceNumber, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventGCStart_V2, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledGCEnd_V1(void)
{
    return ep_event_is_enabled(EventPipeEventGCEnd_V1);
}

uint32_t EventPipeWriteEventGCEnd_V1(
    const uint32_t Count,
    const uint32_t Depth,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledGCEnd_V1())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(Count, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Depth, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventGCEnd_V1, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledBulkType(void)
{
    return ep_event_is_enabled(EventPipeEventBulkType);
}

uint32_t EventPipeWriteEventBulkType(
    const uint32_t Count,
    const uint16_t ClrInstanceID,
    int Values_ElementSize,
    const void* Values,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledBulkType())
        return ERROR_SUCCESS;

    size_t size = 38;
    uint8_t stackBuffer[38];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(Count, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer((const uint8_t *)Values, (int)(Values_ElementSize), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventBulkType, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledGCBulkRootEdge(void)
{
    return ep_event_is_enabled(EventPipeEventGCBulkRootEdge);
}

uint32_t EventPipeWriteEventGCBulkRootEdge(
    const uint32_t Index,
    const uint32_t Count,
    const uint16_t ClrInstanceID,
    int Values_ElementSize,
    const void* Values,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledGCBulkRootEdge())
        return ERROR_SUCCESS;

    size_t size = 42;
    uint8_t stackBuffer[42];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(Index, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Count, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer((const uint8_t *)Values, (int)Values_ElementSize * (int)Count, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventGCBulkRootEdge, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledGCBulkRootConditionalWeakTableElementEdge(void)
{
    return ep_event_is_enabled(EventPipeEventGCBulkRootConditionalWeakTableElementEdge);
}

uint32_t EventPipeWriteEventGCBulkRootConditionalWeakTableElementEdge(
    const uint32_t Index,
    const uint32_t Count,
    const uint16_t ClrInstanceID,
    int Values_ElementSize,
    const void* Values,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledGCBulkRootConditionalWeakTableElementEdge())
        return ERROR_SUCCESS;

    size_t size = 42;
    uint8_t stackBuffer[42];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(Index, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Count, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer((const uint8_t *)Values, (int)Values_ElementSize * (int)Count, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventGCBulkRootConditionalWeakTableElementEdge, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledGCBulkNode(void)
{
    return ep_event_is_enabled(EventPipeEventGCBulkNode);
}

uint32_t EventPipeWriteEventGCBulkNode(
    const uint32_t Index,
    const uint32_t Count,
    const uint16_t ClrInstanceID,
    int Values_ElementSize,
    const void* Values,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledGCBulkNode())
        return ERROR_SUCCESS;

    size_t size = 42;
    uint8_t stackBuffer[42];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(Index, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Count, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer((const uint8_t *)Values, (int)Values_ElementSize * (int)Count, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventGCBulkNode, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledGCBulkEdge(void)
{
    return ep_event_is_enabled(EventPipeEventGCBulkEdge);
}

uint32_t EventPipeWriteEventGCBulkEdge(
    const uint32_t Index,
    const uint32_t Count,
    const uint16_t ClrInstanceID,
    int Values_ElementSize,
    const void* Values,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledGCBulkEdge())
        return ERROR_SUCCESS;

    size_t size = 42;
    uint8_t stackBuffer[42];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(Index, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Count, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer((const uint8_t *)Values, (int)Values_ElementSize * (int)Count, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventGCBulkEdge, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledGCBulkRootStaticVar(void)
{
    return ep_event_is_enabled(EventPipeEventGCBulkRootStaticVar);
}

uint32_t EventPipeWriteEventGCBulkRootStaticVar(
    const uint32_t Count,
    const uint64_t AppDomainID,
    const uint16_t ClrInstanceID,
    int Values_ElementSize,
    const void* Values,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledGCBulkRootStaticVar())
        return ERROR_SUCCESS;

    size_t size = 46;
    uint8_t stackBuffer[46];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(Count, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer((const uint8_t *)Values, (int)(Values_ElementSize), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventGCBulkRootStaticVar, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledThreadPoolWorkerThreadStart(void)
{
    return ep_event_is_enabled(EventPipeEventThreadPoolWorkerThreadStart);
}

uint32_t EventPipeWriteEventThreadPoolWorkerThreadStart(
    const uint32_t ActiveWorkerThreadCount,
    const uint32_t RetiredWorkerThreadCount,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledThreadPoolWorkerThreadStart())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(ActiveWorkerThreadCount, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(RetiredWorkerThreadCount, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventThreadPoolWorkerThreadStart, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledThreadPoolWorkerThreadStop(void)
{
    return ep_event_is_enabled(EventPipeEventThreadPoolWorkerThreadStop);
}

uint32_t EventPipeWriteEventThreadPoolWorkerThreadStop(
    const uint32_t ActiveWorkerThreadCount,
    const uint32_t RetiredWorkerThreadCount,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledThreadPoolWorkerThreadStop())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(ActiveWorkerThreadCount, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(RetiredWorkerThreadCount, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventThreadPoolWorkerThreadStop, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledThreadPoolWorkerThreadAdjustmentSample(void)
{
    return ep_event_is_enabled(EventPipeEventThreadPoolWorkerThreadAdjustmentSample);
}

uint32_t EventPipeWriteEventThreadPoolWorkerThreadAdjustmentSample(
    const double Throughput,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledThreadPoolWorkerThreadAdjustmentSample())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_double_t(Throughput, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventThreadPoolWorkerThreadAdjustmentSample, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledThreadPoolWorkerThreadAdjustmentAdjustment(void)
{
    return ep_event_is_enabled(EventPipeEventThreadPoolWorkerThreadAdjustmentAdjustment);
}

uint32_t EventPipeWriteEventThreadPoolWorkerThreadAdjustmentAdjustment(
    const double AverageThroughput,
    const uint32_t NewWorkerThreadCount,
    const uint32_t Reason,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledThreadPoolWorkerThreadAdjustmentAdjustment())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_double_t(AverageThroughput, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(NewWorkerThreadCount, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Reason, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventThreadPoolWorkerThreadAdjustmentAdjustment, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledThreadPoolWorkerThreadAdjustmentStats(void)
{
    return ep_event_is_enabled(EventPipeEventThreadPoolWorkerThreadAdjustmentStats);
}

uint32_t EventPipeWriteEventThreadPoolWorkerThreadAdjustmentStats(
    const double Duration,
    const double Throughput,
    const double ThreadWave,
    const double ThroughputWave,
    const double ThroughputErrorEstimate,
    const double AverageThroughputErrorEstimate,
    const double ThroughputRatio,
    const double Confidence,
    const double NewControlSetting,
    const uint16_t NewThreadWaveMagnitude,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledThreadPoolWorkerThreadAdjustmentStats())
        return ERROR_SUCCESS;

    size_t size = 76;
    uint8_t stackBuffer[76];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_double_t(Duration, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_double_t(Throughput, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_double_t(ThreadWave, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_double_t(ThroughputWave, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_double_t(ThroughputErrorEstimate, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_double_t(AverageThroughputErrorEstimate, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_double_t(ThroughputRatio, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_double_t(Confidence, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_double_t(NewControlSetting, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(NewThreadWaveMagnitude, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventThreadPoolWorkerThreadAdjustmentStats, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledThreadPoolWorkerThreadWait(void)
{
    return ep_event_is_enabled(EventPipeEventThreadPoolWorkerThreadWait);
}

uint32_t EventPipeWriteEventThreadPoolWorkerThreadWait(
    const uint32_t ActiveWorkerThreadCount,
    const uint32_t RetiredWorkerThreadCount,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledThreadPoolWorkerThreadWait())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(ActiveWorkerThreadCount, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(RetiredWorkerThreadCount, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventThreadPoolWorkerThreadWait, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledThreadPoolMinMaxThreads(void)
{
    return ep_event_is_enabled(EventPipeEventThreadPoolMinMaxThreads);
}

uint32_t EventPipeWriteEventThreadPoolMinMaxThreads(
    const uint16_t MinWorkerThreads,
    const uint16_t MaxWorkerThreads,
    const uint16_t MinIOCompletionThreads,
    const uint16_t MaxIOCompletionThreads,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledThreadPoolMinMaxThreads())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint16_t(MinWorkerThreads, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(MaxWorkerThreads, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(MinIOCompletionThreads, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(MaxIOCompletionThreads, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventThreadPoolMinMaxThreads, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledThreadPoolWorkingThreadCount(void)
{
    return ep_event_is_enabled(EventPipeEventThreadPoolWorkingThreadCount);
}

uint32_t EventPipeWriteEventThreadPoolWorkingThreadCount(
    const uint32_t Count,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledThreadPoolWorkingThreadCount())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(Count, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventThreadPoolWorkingThreadCount, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledThreadPoolIOEnqueue(void)
{
    return ep_event_is_enabled(EventPipeEventThreadPoolIOEnqueue);
}

uint32_t EventPipeWriteEventThreadPoolIOEnqueue(
    const void* NativeOverlapped,
    const void* Overlapped,
    const bool MultiDequeues,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledThreadPoolIOEnqueue())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)NativeOverlapped, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uintptr_t((uintptr_t)Overlapped, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_bool_t(MultiDequeues, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventThreadPoolIOEnqueue, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledThreadPoolIODequeue(void)
{
    return ep_event_is_enabled(EventPipeEventThreadPoolIODequeue);
}

uint32_t EventPipeWriteEventThreadPoolIODequeue(
    const void* NativeOverlapped,
    const void* Overlapped,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledThreadPoolIODequeue())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)NativeOverlapped, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uintptr_t((uintptr_t)Overlapped, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventThreadPoolIODequeue, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledThreadPoolIOPack(void)
{
    return ep_event_is_enabled(EventPipeEventThreadPoolIOPack);
}

uint32_t EventPipeWriteEventThreadPoolIOPack(
    const void* NativeOverlapped,
    const void* Overlapped,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledThreadPoolIOPack())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)NativeOverlapped, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uintptr_t((uintptr_t)Overlapped, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventThreadPoolIOPack, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMethodDetails(void)
{
    return ep_event_is_enabled(EventPipeEventMethodDetails);
}

uint32_t EventPipeWriteEventMethodDetails(
    const uint64_t MethodID,
    const uint64_t TypeID,
    const uint32_t MethodToken,
    const uint32_t TypeParameterCount,
    const uint64_t LoaderModuleID,
    const uint64_t* TypeParameters,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMethodDetails())
        return ERROR_SUCCESS;

    size_t size = 40;
    uint8_t stackBuffer[40];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(TypeID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(MethodToken, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(TypeParameterCount, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(LoaderModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer((const uint8_t *)TypeParameters, sizeof(const uint64_t) * (int)TypeParameterCount, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMethodDetails, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledTypeLoadStart(void)
{
    return ep_event_is_enabled(EventPipeEventTypeLoadStart);
}

uint32_t EventPipeWriteEventTypeLoadStart(
    const uint32_t TypeLoadStartID,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledTypeLoadStart())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint32_t(TypeLoadStartID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventTypeLoadStart, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledTypeLoadStop(void)
{
    return ep_event_is_enabled(EventPipeEventTypeLoadStop);
}

uint32_t EventPipeWriteEventTypeLoadStop(
    const uint32_t TypeLoadStartID,
    const uint16_t ClrInstanceID,
    const uint16_t LoadLevel,
    const uint64_t TypeID,
    const ep_char8_t* TypeName,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledTypeLoadStop())
        return ERROR_SUCCESS;

    size_t size = 80;
    uint8_t stackBuffer[80];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!TypeName) { TypeName = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint32_t(TypeLoadStartID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(LoadLevel, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(TypeID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(TypeName, strlen((const char *)TypeName), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventTypeLoadStop, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledExceptionThrown_V1(void)
{
    return ep_event_is_enabled(EventPipeEventExceptionThrown_V1);
}

uint32_t EventPipeWriteEventExceptionThrown_V1(
    const ep_char8_t* ExceptionType,
    const ep_char8_t* ExceptionMessage,
    const void* ExceptionEIP,
    const uint32_t ExceptionHRESULT,
    const uint16_t ExceptionFlags,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledExceptionThrown_V1())
        return ERROR_SUCCESS;

    size_t size = 144;
    uint8_t stackBuffer[144];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!ExceptionType) { ExceptionType = (const ep_char8_t *)"NULL"; }
    if (!ExceptionMessage) { ExceptionMessage = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_string_utf8_to_utf16_t(ExceptionType, strlen((const char *)ExceptionType), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ExceptionMessage, strlen((const char *)ExceptionMessage), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uintptr_t((uintptr_t)ExceptionEIP, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(ExceptionHRESULT, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ExceptionFlags, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventExceptionThrown_V1, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledExceptionCatchStart(void)
{
    return ep_event_is_enabled(EventPipeEventExceptionCatchStart);
}

uint32_t EventPipeWriteEventExceptionCatchStart(
    const uint64_t EntryEIP,
    const uint64_t MethodID,
    const ep_char8_t* MethodName,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledExceptionCatchStart())
        return ERROR_SUCCESS;

    size_t size = 82;
    uint8_t stackBuffer[82];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!MethodName) { MethodName = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(EntryEIP, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(MethodName, strlen((const char *)MethodName), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventExceptionCatchStart, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledExceptionCatchStop(void)
{
    return ep_event_is_enabled(EventPipeEventExceptionCatchStop);
}

uint32_t EventPipeWriteEventExceptionCatchStop(    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledExceptionCatchStop())
        return ERROR_SUCCESS;
    ep_write_event (EventPipeEventExceptionCatchStop, (uint8_t *) NULL, 0, (const uint8_t *)ActivityId, (const uint8_t *)RelatedActivityId);

    return ERROR_SUCCESS;
}

bool EventPipeEventEnabledExceptionFinallyStart(void)
{
    return ep_event_is_enabled(EventPipeEventExceptionFinallyStart);
}

uint32_t EventPipeWriteEventExceptionFinallyStart(
    const uint64_t EntryEIP,
    const uint64_t MethodID,
    const ep_char8_t* MethodName,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledExceptionFinallyStart())
        return ERROR_SUCCESS;

    size_t size = 82;
    uint8_t stackBuffer[82];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!MethodName) { MethodName = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(EntryEIP, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(MethodName, strlen((const char *)MethodName), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventExceptionFinallyStart, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledExceptionFinallyStop(void)
{
    return ep_event_is_enabled(EventPipeEventExceptionFinallyStop);
}

uint32_t EventPipeWriteEventExceptionFinallyStop(    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledExceptionFinallyStop())
        return ERROR_SUCCESS;
    ep_write_event (EventPipeEventExceptionFinallyStop, (uint8_t *) NULL, 0, (const uint8_t *)ActivityId, (const uint8_t *)RelatedActivityId);

    return ERROR_SUCCESS;
}

bool EventPipeEventEnabledExceptionFilterStart(void)
{
    return ep_event_is_enabled(EventPipeEventExceptionFilterStart);
}

uint32_t EventPipeWriteEventExceptionFilterStart(
    const uint64_t EntryEIP,
    const uint64_t MethodID,
    const ep_char8_t* MethodName,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledExceptionFilterStart())
        return ERROR_SUCCESS;

    size_t size = 82;
    uint8_t stackBuffer[82];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!MethodName) { MethodName = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(EntryEIP, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(MethodName, strlen((const char *)MethodName), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventExceptionFilterStart, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledExceptionFilterStop(void)
{
    return ep_event_is_enabled(EventPipeEventExceptionFilterStop);
}

uint32_t EventPipeWriteEventExceptionFilterStop(    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledExceptionFilterStop())
        return ERROR_SUCCESS;
    ep_write_event (EventPipeEventExceptionFilterStop, (uint8_t *) NULL, 0, (const uint8_t *)ActivityId, (const uint8_t *)RelatedActivityId);

    return ERROR_SUCCESS;
}

bool EventPipeEventEnabledExceptionThrownStop(void)
{
    return ep_event_is_enabled(EventPipeEventExceptionThrownStop);
}

uint32_t EventPipeWriteEventExceptionThrownStop(    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledExceptionThrownStop())
        return ERROR_SUCCESS;
    ep_write_event (EventPipeEventExceptionThrownStop, (uint8_t *) NULL, 0, (const uint8_t *)ActivityId, (const uint8_t *)RelatedActivityId);

    return ERROR_SUCCESS;
}

bool EventPipeEventEnabledContentionStart_V1(void)
{
    return ep_event_is_enabled(EventPipeEventContentionStart_V1);
}

uint32_t EventPipeWriteEventContentionStart_V1(
    const uint8_t ContentionFlags,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledContentionStart_V1())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint8_t(ContentionFlags, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventContentionStart_V1, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledContentionStart_V2(void)
{
    return ep_event_is_enabled(EventPipeEventContentionStart_V2);
}

uint32_t EventPipeWriteEventContentionStart_V2(
    const uint8_t ContentionFlags,
    const uint16_t ClrInstanceID,
    const void* LockID,
    const void* AssociatedObjectID,
    const uint64_t LockOwnerThreadID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledContentionStart_V2())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint8_t(ContentionFlags, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uintptr_t((uintptr_t)LockID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uintptr_t((uintptr_t)AssociatedObjectID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(LockOwnerThreadID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventContentionStart_V2, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledContentionStop(void)
{
    return ep_event_is_enabled(EventPipeEventContentionStop);
}

uint32_t EventPipeWriteEventContentionStop(
    const uint8_t ContentionFlags,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledContentionStop())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint8_t(ContentionFlags, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventContentionStop, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledContentionStop_V1(void)
{
    return ep_event_is_enabled(EventPipeEventContentionStop_V1);
}

uint32_t EventPipeWriteEventContentionStop_V1(
    const uint8_t ContentionFlags,
    const uint16_t ClrInstanceID,
    const double DurationNs,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledContentionStop_V1())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint8_t(ContentionFlags, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_double_t(DurationNs, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventContentionStop_V1, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledContentionLockCreated(void)
{
    return ep_event_is_enabled(EventPipeEventContentionLockCreated);
}

uint32_t EventPipeWriteEventContentionLockCreated(
    const void* LockID,
    const void* AssociatedObjectID,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledContentionLockCreated())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uintptr_t((uintptr_t)LockID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uintptr_t((uintptr_t)AssociatedObjectID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventContentionLockCreated, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledThreadCreated(void)
{
    return ep_event_is_enabled(EventPipeEventThreadCreated);
}

uint32_t EventPipeWriteEventThreadCreated(
    const uint64_t ManagedThreadID,
    const uint64_t AppDomainID,
    const uint32_t Flags,
    const uint32_t ManagedThreadIndex,
    const uint32_t OSThreadID,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledThreadCreated())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(ManagedThreadID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Flags, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(ManagedThreadIndex, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(OSThreadID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventThreadCreated, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledThreadTerminated(void)
{
    return ep_event_is_enabled(EventPipeEventThreadTerminated);
}

uint32_t EventPipeWriteEventThreadTerminated(
    const uint64_t ManagedThreadID,
    const uint64_t AppDomainID,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledThreadTerminated())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(ManagedThreadID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventThreadTerminated, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMethodLoad_V1(void)
{
    return ep_event_is_enabled(EventPipeEventMethodLoad_V1);
}

uint32_t EventPipeWriteEventMethodLoad_V1(
    const uint64_t MethodID,
    const uint64_t ModuleID,
    const uint64_t MethodStartAddress,
    const uint32_t MethodSize,
    const uint32_t MethodToken,
    const uint32_t MethodFlags,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMethodLoad_V1())
        return ERROR_SUCCESS;

    size_t size = 38;
    uint8_t stackBuffer[38];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(MethodStartAddress, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(MethodSize, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(MethodToken, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(MethodFlags, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMethodLoad_V1, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMethodLoadVerbose_V1(void)
{
    return ep_event_is_enabled(EventPipeEventMethodLoadVerbose_V1);
}

uint32_t EventPipeWriteEventMethodLoadVerbose_V1(
    const uint64_t MethodID,
    const uint64_t ModuleID,
    const uint64_t MethodStartAddress,
    const uint32_t MethodSize,
    const uint32_t MethodToken,
    const uint32_t MethodFlags,
    const ep_char8_t* MethodNamespace,
    const ep_char8_t* MethodName,
    const ep_char8_t* MethodSignature,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMethodLoadVerbose_V1())
        return ERROR_SUCCESS;

    size_t size = 230;
    uint8_t stackBuffer[230];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!MethodNamespace) { MethodNamespace = (const ep_char8_t *)"NULL"; }
    if (!MethodName) { MethodName = (const ep_char8_t *)"NULL"; }
    if (!MethodSignature) { MethodSignature = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(MethodStartAddress, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(MethodSize, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(MethodToken, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(MethodFlags, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(MethodNamespace, strlen((const char *)MethodNamespace), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(MethodName, strlen((const char *)MethodName), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(MethodSignature, strlen((const char *)MethodSignature), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMethodLoadVerbose_V1, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMethodJittingStarted_V1(void)
{
    return ep_event_is_enabled(EventPipeEventMethodJittingStarted_V1);
}

uint32_t EventPipeWriteEventMethodJittingStarted_V1(
    const uint64_t MethodID,
    const uint64_t ModuleID,
    const uint32_t MethodToken,
    const uint32_t MethodILSize,
    const ep_char8_t* MethodNamespace,
    const ep_char8_t* MethodName,
    const ep_char8_t* MethodSignature,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMethodJittingStarted_V1())
        return ERROR_SUCCESS;

    size_t size = 218;
    uint8_t stackBuffer[218];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!MethodNamespace) { MethodNamespace = (const ep_char8_t *)"NULL"; }
    if (!MethodName) { MethodName = (const ep_char8_t *)"NULL"; }
    if (!MethodSignature) { MethodSignature = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(MethodToken, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(MethodILSize, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(MethodNamespace, strlen((const char *)MethodNamespace), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(MethodName, strlen((const char *)MethodName), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(MethodSignature, strlen((const char *)MethodSignature), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMethodJittingStarted_V1, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMethodJitMemoryAllocatedForCode(void)
{
    return ep_event_is_enabled(EventPipeEventMethodJitMemoryAllocatedForCode);
}

uint32_t EventPipeWriteEventMethodJitMemoryAllocatedForCode(
    const uint64_t MethodID,
    const uint64_t ModuleID,
    const uint64_t JitHotCodeRequestSize,
    const uint64_t JitRODataRequestSize,
    const uint64_t AllocatedSizeForJitCode,
    const uint32_t JitAllocFlag,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMethodJitMemoryAllocatedForCode())
        return ERROR_SUCCESS;

    size_t size = 46;
    uint8_t stackBuffer[46];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(JitHotCodeRequestSize, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(JitRODataRequestSize, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AllocatedSizeForJitCode, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(JitAllocFlag, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMethodJitMemoryAllocatedForCode, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledMethodILToNativeMap(void)
{
    return ep_event_is_enabled(EventPipeEventMethodILToNativeMap);
}

uint32_t EventPipeWriteEventMethodILToNativeMap(
    const uint64_t MethodID,
    const uint64_t ReJITID,
    const uint8_t MethodExtent,
    const uint16_t CountOfMapEntries,
    const uint32_t* ILOffsets,
    const uint32_t* NativeOffsets,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledMethodILToNativeMap())
        return ERROR_SUCCESS;

    size_t size = 32;
    uint8_t stackBuffer[32];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    success &= write_buffer_uint64_t(MethodID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(ReJITID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint8_t(MethodExtent, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(CountOfMapEntries, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer((const uint8_t *)ILOffsets, sizeof(const uint32_t) * (int)CountOfMapEntries, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer((const uint8_t *)NativeOffsets, sizeof(const uint32_t) * (int)CountOfMapEntries, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventMethodILToNativeMap, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledDomainModuleLoad_V1(void)
{
    return ep_event_is_enabled(EventPipeEventDomainModuleLoad_V1);
}

uint32_t EventPipeWriteEventDomainModuleLoad_V1(
    const uint64_t ModuleID,
    const uint64_t AssemblyID,
    const uint64_t AppDomainID,
    const uint32_t ModuleFlags,
    const uint32_t Reserved1,
    const ep_char8_t* ModuleILPath,
    const ep_char8_t* ModuleNativePath,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledDomainModuleLoad_V1())
        return ERROR_SUCCESS;

    size_t size = 162;
    uint8_t stackBuffer[162];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!ModuleILPath) { ModuleILPath = (const ep_char8_t *)"NULL"; }
    if (!ModuleNativePath) { ModuleNativePath = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AssemblyID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(ModuleFlags, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Reserved1, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ModuleILPath, strlen((const char *)ModuleILPath), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ModuleNativePath, strlen((const char *)ModuleNativePath), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventDomainModuleLoad_V1, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledModuleLoad_V2(void)
{
    return ep_event_is_enabled(EventPipeEventModuleLoad_V2);
}

uint32_t EventPipeWriteEventModuleLoad_V2(
    const uint64_t ModuleID,
    const uint64_t AssemblyID,
    const uint32_t ModuleFlags,
    const uint32_t Reserved1,
    const ep_char8_t* ModuleILPath,
    const ep_char8_t* ModuleNativePath,
    const uint16_t ClrInstanceID,
    const uint8_t* ManagedPdbSignature,
    const uint32_t ManagedPdbAge,
    const ep_char8_t* ManagedPdbBuildPath,
    const uint8_t* NativePdbSignature,
    const uint32_t NativePdbAge,
    const ep_char8_t* NativePdbBuildPath,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledModuleLoad_V2())
        return ERROR_SUCCESS;

    size_t size = 290;
    uint8_t stackBuffer[290];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!ModuleILPath) { ModuleILPath = (const ep_char8_t *)"NULL"; }
    if (!ModuleNativePath) { ModuleNativePath = (const ep_char8_t *)"NULL"; }
    if (!ManagedPdbBuildPath) { ManagedPdbBuildPath = (const ep_char8_t *)"NULL"; }
    if (!NativePdbBuildPath) { NativePdbBuildPath = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AssemblyID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(ModuleFlags, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Reserved1, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ModuleILPath, strlen((const char *)ModuleILPath), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ModuleNativePath, strlen((const char *)ModuleNativePath), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_guid_t(ManagedPdbSignature, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(ManagedPdbAge, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ManagedPdbBuildPath, strlen((const char *)ManagedPdbBuildPath), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_guid_t(NativePdbSignature, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(NativePdbAge, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(NativePdbBuildPath, strlen((const char *)NativePdbBuildPath), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventModuleLoad_V2, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledModuleUnload_V2(void)
{
    return ep_event_is_enabled(EventPipeEventModuleUnload_V2);
}

uint32_t EventPipeWriteEventModuleUnload_V2(
    const uint64_t ModuleID,
    const uint64_t AssemblyID,
    const uint32_t ModuleFlags,
    const uint32_t Reserved1,
    const ep_char8_t* ModuleILPath,
    const ep_char8_t* ModuleNativePath,
    const uint16_t ClrInstanceID,
    const uint8_t* ManagedPdbSignature,
    const uint32_t ManagedPdbAge,
    const ep_char8_t* ManagedPdbBuildPath,
    const uint8_t* NativePdbSignature,
    const uint32_t NativePdbAge,
    const ep_char8_t* NativePdbBuildPath,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledModuleUnload_V2())
        return ERROR_SUCCESS;

    size_t size = 290;
    uint8_t stackBuffer[290];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!ModuleILPath) { ModuleILPath = (const ep_char8_t *)"NULL"; }
    if (!ModuleNativePath) { ModuleNativePath = (const ep_char8_t *)"NULL"; }
    if (!ManagedPdbBuildPath) { ManagedPdbBuildPath = (const ep_char8_t *)"NULL"; }
    if (!NativePdbBuildPath) { NativePdbBuildPath = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(ModuleID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AssemblyID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(ModuleFlags, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(Reserved1, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ModuleILPath, strlen((const char *)ModuleILPath), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ModuleNativePath, strlen((const char *)ModuleNativePath), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_guid_t(ManagedPdbSignature, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(ManagedPdbAge, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(ManagedPdbBuildPath, strlen((const char *)ManagedPdbBuildPath), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_guid_t(NativePdbSignature, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(NativePdbAge, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(NativePdbBuildPath, strlen((const char *)NativePdbBuildPath), &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventModuleUnload_V2, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledAssemblyLoad_V1(void)
{
    return ep_event_is_enabled(EventPipeEventAssemblyLoad_V1);
}

uint32_t EventPipeWriteEventAssemblyLoad_V1(
    const uint64_t AssemblyID,
    const uint64_t AppDomainID,
    const uint64_t BindingID,
    const uint32_t AssemblyFlags,
    const ep_char8_t* FullyQualifiedAssemblyName,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledAssemblyLoad_V1())
        return ERROR_SUCCESS;

    size_t size = 94;
    uint8_t stackBuffer[94];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!FullyQualifiedAssemblyName) { FullyQualifiedAssemblyName = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(AssemblyID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(BindingID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(AssemblyFlags, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(FullyQualifiedAssemblyName, strlen((const char *)FullyQualifiedAssemblyName), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventAssemblyLoad_V1, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

bool EventPipeEventEnabledAssemblyUnload_V1(void)
{
    return ep_event_is_enabled(EventPipeEventAssemblyUnload_V1);
}

uint32_t EventPipeWriteEventAssemblyUnload_V1(
    const uint64_t AssemblyID,
    const uint64_t AppDomainID,
    const uint64_t BindingID,
    const uint32_t AssemblyFlags,
    const ep_char8_t* FullyQualifiedAssemblyName,
    const uint16_t ClrInstanceID,
    const uint8_t* ActivityId,
    const uint8_t* RelatedActivityId)
{
    if (!EventPipeEventEnabledAssemblyUnload_V1())
        return ERROR_SUCCESS;

    size_t size = 94;
    uint8_t stackBuffer[94];
    uint8_t *buffer = stackBuffer;
    size_t offset = 0;
    bool fixedBuffer = true;
    bool success = true;

    if (!FullyQualifiedAssemblyName) { FullyQualifiedAssemblyName = (const ep_char8_t *)"NULL"; }
    success &= write_buffer_uint64_t(AssemblyID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(AppDomainID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint64_t(BindingID, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint32_t(AssemblyFlags, &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_string_utf8_to_utf16_t(FullyQualifiedAssemblyName, strlen((const char *)FullyQualifiedAssemblyName), &buffer, &offset, &size, &fixedBuffer);
    success &= write_buffer_uint16_t(ClrInstanceID, &buffer, &offset, &size, &fixedBuffer);

    ep_raise_error_if_nok (success);

    ep_write_event (EventPipeEventAssemblyUnload_V1, (uint8_t *)buffer, (uint32_t)offset, ActivityId, RelatedActivityId);

ep_on_exit:
    if (!fixedBuffer)
        ep_rt_byte_array_free (buffer);
    return success ? ERROR_SUCCESS : ERROR_WRITE_FAULT;

ep_on_error:
    EP_ASSERT (!success);
    ep_exit_error_handler ();
}

void EP_CALLBACK_CALLTYPE EventPipeEtwCallbackDotNETRuntime(const uint8_t *, unsigned long, uint8_t, uint64_t, uint64_t, EventFilterDescriptor *, void *);

void InitDotNETRuntime(void);

void InitDotNETRuntime(void)
{
    EventPipeProviderDotNETRuntime = create_provider(SL(DotNETRuntimeName), (EventPipeEtwCallbackDotNETRuntime));
    EventPipeEventGCStart_V2 = provider_add_event(EventPipeProviderDotNETRuntime,1,1,2,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventGCEnd_V1 = provider_add_event(EventPipeProviderDotNETRuntime,2,1,1,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventBulkType = provider_add_event(EventPipeProviderDotNETRuntime,15,524288,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventGCBulkRootEdge = provider_add_event(EventPipeProviderDotNETRuntime,16,1048576,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventGCBulkRootConditionalWeakTableElementEdge = provider_add_event(EventPipeProviderDotNETRuntime,17,1048576,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventGCBulkNode = provider_add_event(EventPipeProviderDotNETRuntime,18,1048576,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventGCBulkEdge = provider_add_event(EventPipeProviderDotNETRuntime,19,1048576,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventGCBulkRootStaticVar = provider_add_event(EventPipeProviderDotNETRuntime,38,1048576,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventThreadPoolWorkerThreadStart = provider_add_event(EventPipeProviderDotNETRuntime,50,65536,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventThreadPoolWorkerThreadStop = provider_add_event(EventPipeProviderDotNETRuntime,51,65536,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventThreadPoolWorkerThreadAdjustmentSample = provider_add_event(EventPipeProviderDotNETRuntime,54,65536,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventThreadPoolWorkerThreadAdjustmentAdjustment = provider_add_event(EventPipeProviderDotNETRuntime,55,65536,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventThreadPoolWorkerThreadAdjustmentStats = provider_add_event(EventPipeProviderDotNETRuntime,56,65536,0,EP_EVENT_LEVEL_VERBOSE,true);
    EventPipeEventThreadPoolWorkerThreadWait = provider_add_event(EventPipeProviderDotNETRuntime,57,65536,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventThreadPoolMinMaxThreads = provider_add_event(EventPipeProviderDotNETRuntime,59,65536,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventThreadPoolWorkingThreadCount = provider_add_event(EventPipeProviderDotNETRuntime,60,65536,0,EP_EVENT_LEVEL_VERBOSE,true);
    EventPipeEventThreadPoolIOEnqueue = provider_add_event(EventPipeProviderDotNETRuntime,63,2147549184,0,EP_EVENT_LEVEL_VERBOSE,true);
    EventPipeEventThreadPoolIODequeue = provider_add_event(EventPipeProviderDotNETRuntime,64,2147549184,0,EP_EVENT_LEVEL_VERBOSE,true);
    EventPipeEventThreadPoolIOPack = provider_add_event(EventPipeProviderDotNETRuntime,65,65536,0,EP_EVENT_LEVEL_VERBOSE,true);
    EventPipeEventMethodDetails = provider_add_event(EventPipeProviderDotNETRuntime,72,274877906944,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventTypeLoadStart = provider_add_event(EventPipeProviderDotNETRuntime,73,549755813888,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventTypeLoadStop = provider_add_event(EventPipeProviderDotNETRuntime,74,549755813888,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventExceptionThrown_V1 = provider_add_event(EventPipeProviderDotNETRuntime,80,8589967360,1,EP_EVENT_LEVEL_ERROR,true);
    EventPipeEventExceptionCatchStart = provider_add_event(EventPipeProviderDotNETRuntime,250,32768,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventExceptionCatchStop = provider_add_event(EventPipeProviderDotNETRuntime,251,32768,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventExceptionFinallyStart = provider_add_event(EventPipeProviderDotNETRuntime,252,32768,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventExceptionFinallyStop = provider_add_event(EventPipeProviderDotNETRuntime,253,32768,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventExceptionFilterStart = provider_add_event(EventPipeProviderDotNETRuntime,254,32768,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventExceptionFilterStop = provider_add_event(EventPipeProviderDotNETRuntime,255,32768,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventExceptionThrownStop = provider_add_event(EventPipeProviderDotNETRuntime,256,32768,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventContentionStart_V1 = provider_add_event(EventPipeProviderDotNETRuntime,81,16384,1,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventContentionStart_V2 = provider_add_event(EventPipeProviderDotNETRuntime,81,16384,2,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventContentionStop = provider_add_event(EventPipeProviderDotNETRuntime,91,16384,0,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventContentionStop_V1 = provider_add_event(EventPipeProviderDotNETRuntime,91,16384,1,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventContentionLockCreated = provider_add_event(EventPipeProviderDotNETRuntime,90,16384,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventThreadCreated = provider_add_event(EventPipeProviderDotNETRuntime,85,67584,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventThreadTerminated = provider_add_event(EventPipeProviderDotNETRuntime,86,67584,0,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventMethodLoad_V1 = provider_add_event(EventPipeProviderDotNETRuntime,141,48,1,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMethodLoadVerbose_V1 = provider_add_event(EventPipeProviderDotNETRuntime,143,48,1,EP_EVENT_LEVEL_INFORMATIONAL,false);
    EventPipeEventMethodJittingStarted_V1 = provider_add_event(EventPipeProviderDotNETRuntime,145,16,1,EP_EVENT_LEVEL_VERBOSE,true);
    EventPipeEventMethodJitMemoryAllocatedForCode = provider_add_event(EventPipeProviderDotNETRuntime,146,16,0,EP_EVENT_LEVEL_VERBOSE,true);
    EventPipeEventMethodILToNativeMap = provider_add_event(EventPipeProviderDotNETRuntime,190,131072,0,EP_EVENT_LEVEL_VERBOSE,true);
    EventPipeEventDomainModuleLoad_V1 = provider_add_event(EventPipeProviderDotNETRuntime,151,8,1,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventModuleLoad_V2 = provider_add_event(EventPipeProviderDotNETRuntime,152,536870920,2,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventModuleUnload_V2 = provider_add_event(EventPipeProviderDotNETRuntime,153,536870920,2,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventAssemblyLoad_V1 = provider_add_event(EventPipeProviderDotNETRuntime,154,8,1,EP_EVENT_LEVEL_INFORMATIONAL,true);
    EventPipeEventAssemblyUnload_V1 = provider_add_event(EventPipeProviderDotNETRuntime,155,8,1,EP_EVENT_LEVEL_INFORMATIONAL,true);
}
#endif
